#define _OPC_DATA_C

#include <assert.h>

#include <opcerr.h>
#include <opcerror.h>
#include <opcdebug.h>
#include <opcdata.h>

#define OPC_DATA_ALLOCATE_PRESENTATIONS 2

static inline void opcDataZeroContext(OPCDataContext ctx) {
    assert(ctx);

    ctx->representation = NULL;
    ctx->native_representation = NULL;
    ctx->opcxml_representation = NULL;
    
    ctx->chunks = NULL;
    ctx->current = NULL;
    ctx->lock = NULL;
    ctx->representation_lock = NULL;
    ctx->chunk_lock = NULL;
}

OPCDataContext opcDataCreateContext(OPCMallocContext mctx) {
    OPCDataContext ctx;

    ctx = (OPCDataContext)opcMalloc(mctx,sizeof(struct OPCDataContextT));
    opcDebugServer(OPC_DEBUG_NOTICE, "opcDataCreateContext: %p", ctx);    
    if (ctx) {
	ctx->malloc=mctx;
	opcDataZeroContext(ctx);
    }
    return ctx;
}

int opcDataInitContext(OPCDataContext ctx, OPCContext sctx, OPCType type) {
    int err;
    OPCBufferPoolContext buffer_pool;
    OPCChunkPoolContext chunk_pool;
    OPCDataRepresentationContext rctx;
    OPCFormatsContext fctx;
    OPCAllocatorContext allocator;

    opcDebugServer(OPC_DEBUG_NOTICE, "opcDataInitContext: %p (Type: %p)", ctx, type);

    assert(ctx);
    assert(sctx);
    assert(type);

    if ((!ctx)||(!sctx)||(!type)) return ERR_OPC_PARAM;

    fctx = sctx->formats;
    chunk_pool = sctx->chunks;
    buffer_pool = sctx->buffers;

    if ((!fctx)||(!chunk_pool)||(!buffer_pool))
	return ERR_OPC_DATA_INCOMPLETE_SERVER_CONTEXT;
    
    if ((!fctx->native_format)||(!fctx->opcxml_format))
	return ERR_OPC_DATA_INCOMPLETE_FORMATS_CONTEXT;
	
    opcDataZeroContext(ctx);    

    ctx->allocated_representations = OPC_DATA_ALLOCATE_PRESENTATIONS;

    ctx->representation = (OPCDataRepresentationContext*)opcMalloc(ctx->malloc, ctx->allocated_representations*sizeof(OPCDataRepresentationContext));
    if (!ctx->representation) return ERR_OPC_MEM;

    ctx->type = type;
    ctx->chunk_pool = chunk_pool;
    ctx->buffer_pool = buffer_pool;

    allocator = opcDataCreateDefaultAllocator(ctx, fctx->native_format, NULL); /* ToDo: options */
    if ((!allocator)&&(fctx->native_format->allocator != OPC_ALLOCATOR_TYPE_ABSENT)) {
	opcDataFreeContext(ctx);
	return ERR_OPC_DATA_CREATE_ALLOCATOR_FOR_NATIVE_REPRESENTATION;
    }
	
    rctx=opcDataRepresentationCreateContext(ctx->malloc);
    if (rctx) {
	err = opcDataRepresentationInitContext(rctx, fctx->native_format, ctx->type, allocator); 
	if (err) {
	    if (allocator) opcAllocatorDestroyContext(allocator);
	    opcDataRepresentationDestroyContext(rctx);
	    opcDataFreeContext(ctx);
	    return err;
	}
	opcDataRepresentationSetId(rctx, OPC_REPRESENTATION_NATIVE);
    } else {
	if (allocator) opcAllocatorDestroyContext(allocator);
	opcDataFreeContext(ctx);
	return ERR_OPC_DATA_CREATE_NATIVE_REPRESENTATION;
    }
    
    ctx->native_representation = rctx;
    ctx->representation[OPC_REPRESENTATION_NATIVE] = rctx;
    

    allocator = opcDataCreateDefaultAllocator(ctx, fctx->opcxml_format, NULL); /* ToDo: options */
    if ((!allocator)&&(fctx->opcxml_format->allocator !=OPC_ALLOCATOR_TYPE_ABSENT)) {
	opcDataFreeContext(ctx);
	return ERR_OPC_DATA_CREATE_ALLOCATOR_FOR_OPCXML_REPRESENTATION;
    }

    rctx=opcDataRepresentationCreateContext(ctx->malloc);
    if (rctx) {
	err = opcDataRepresentationInitContext(rctx, fctx->opcxml_format, ctx->type, allocator); 
	if (err) {
	    if (allocator) opcAllocatorDestroyContext(allocator);
	    opcDataRepresentationDestroyContext(rctx);
	    opcDataFreeContext(ctx);
	    return err;
	}
	opcDataRepresentationSetId(rctx, OPC_REPRESENTATION_OPCXML);
	opcDataRepresentationSetBaseRepresentation(rctx, ctx->native_representation);
    } else {
	if (allocator) opcAllocatorDestroyContext(allocator);
	opcDataFreeContext(ctx);
	return ERR_OPC_DATA_CREATE_OPCXML_REPRESENTATION;
    }
    
    ctx->opcxml_representation = rctx;
    ctx->representation[OPC_REPRESENTATION_OPCXML] = rctx;
    
    ctx->supported_representations = 2;
    
    ctx->lock = LockCreateContext(ctx->malloc);
    ctx->chunk_lock = LockCreateContext(ctx->malloc);
    ctx->representation_lock = LockCreateContext(ctx->malloc);
    if ((!ctx->lock)||(!ctx->chunk_lock)||(!ctx->representation_lock)) {
	opcDataFreeContext(ctx);
	return ERR_OPC_LOCK_CREATE;
    }
    err = LockInitContext(ctx->lock);
    if (!err) err = LockInitContext(ctx->chunk_lock);
    if (!err) err = LockInitContext(ctx->representation_lock);
    if (err) {
	opcDataFreeContext(ctx);
	return err;
    }

    
    ctx->sctx = sctx;
    ctx->fctx = fctx;

    
    return 0;
}

void opcDataFreeContext(OPCDataContext ctx) {
    OPCDataChunkContext chunk, nchunk;
    long i;

    opcDebugServer(OPC_DEBUG_NOTICE, "opcDataFreeContext: %p", ctx);    

    assert(ctx);
    
    if (ctx->lock) LockDestroyContext(ctx->lock);
    if (ctx->chunk_lock) LockDestroyContext(ctx->chunk_lock);
    if (ctx->representation_lock) LockDestroyContext(ctx->representation_lock);

    for (chunk=ctx->chunks;chunk;chunk=nchunk) {
        nchunk = chunk->next;
	// ToDo: Wait until 0 refs
        opcDataChunkDestroyContext(chunk);
    }
    
    if (ctx->representation) {
	for (i=ctx->supported_representations-1;i>=0;i--)
	    opcDataRepresentationDestroyContext(ctx->representation[i]);
	opcFree(ctx->malloc, ctx->representation);
    }
    
    opcDataZeroContext(ctx);    
}

void opcDataDestroyContext(OPCDataContext ctx) {
    opcDebugServer(OPC_DEBUG_NOTICE, "opcDataDestroyContext: %p", ctx);    
    opcDataFreeContext(ctx);
    opcFree(ctx->malloc,ctx);
}

unsigned short int opcDataAddRepresentation(OPCDataContext ctx, OPCFormatContext format, OPCAllocatorContext allocator) {
    int err;
    unsigned short int result;
    OPCDataRepresentationContext rctx;
    OPCDataRepresentationContext *representations;
    
    assert(ctx);
    assert(ctx->type);
    assert(format);
    assert(format->name);

    opcDebugServer(OPC_DEBUG_NOTICE, "opcDataAddFormat: %p. Format: %s.", ctx, format->name);    
    
    if (ctx->supported_representations==ctx->allocated_representations) {
	LockWrite(ctx->representation_lock);
	representations = (OPCDataRepresentationContext*)opcRealloc(ctx->malloc, ctx->representation, 2*ctx->allocated_representations*sizeof(OPCDataRepresentationContext));
	if (!representations) return (unsigned short int)-1;
	ctx->representation = representations;
	ctx->allocated_representations*=2;
	LockReturn(ctx->representation_lock);
    }

	// Creating allocator if required
    if ((!allocator)&&(format->allocator != OPC_ALLOCATOR_TYPE_ABSENT)) {
	opcDebugServer(OPC_DEBUG_ERROR, "opcDataAddFormat: %p. Format: %s. %s",ctx, format->name, opcError(ERR_OPC_DATA_REPRESENTATION_INVALID_ALLOCATOR,NULL));
//	ToDo: Create Default?
//	allocator = opcDataCreateDefaultAllocator(ctx, fctx->opcxml_format, NULL); /* ToDo: options */
	return (unsigned short int)-1;
    }

    
    rctx=opcDataRepresentationCreateContext(ctx->malloc);
    if (rctx) {
	err = opcDataRepresentationInitContext(rctx, format, ctx->type, allocator);
	if (err) {
	    opcDataRepresentationDestroyContext(rctx);
	    opcDebugServer(OPC_DEBUG_ERROR, "opcData(%p): AddRepresentation(%s): Failed: %s.", ctx, format->name, opcError(err, NULL));
	    return (unsigned short int)-1;	    
	}
	opcDataRepresentationSetBaseRepresentation(rctx, ctx->native_representation);
	
	LockWrite(ctx->representation_lock);
	result=ctx->supported_representations++;
	opcDataRepresentationSetId(rctx, result);
	ctx->representation[result]=rctx;
	LockReturn(ctx->representation_lock);
    } else {
	opcDebugServer(OPC_DEBUG_ERROR, "opcData(%p): AddRepresentation(%s): Failed: %s", ctx, format->name, opcError(ERR_OPC_DATA_CREATE_REPRESENTATION, NULL));
	return (unsigned short int)-1;
    }

    opcDebugServer(OPC_DEBUG_DEBUG, "opcData(%p): AddRepresentation(%s): Registered as %u.", ctx, format->name, result);

    return result;
}

int opcDataPush(OPCDataContext ctx, unsigned short int representation, const OPCTime time, const OPCQuality quality, const void *data, unsigned long size) {
    int err;

    OPCDataRepresentationContext rctx;
    OPCDataChunkContext chunk;
    

    assert(ctx);
    assert(data);

    LockRead(ctx->representation_lock);
    assert(ctx->representation);
    rctx = ctx->representation[representation];
    LockReturn(ctx->representation_lock);
    assert(rctx);

    
    opcDebugServer(OPC_DEBUG_DEBUG, "opcData(%p): Push (representation: %u).", ctx, representation);
    
    chunk = opcChunkPoolGetChunk(ctx->chunk_pool, ctx->supported_representations);
    if (!chunk) return ERR_OPC_CHUNKPOOL_GET_CHUNK;

    opcDataChunkSetRepresentationPtr(chunk, ctx->representation);
    opcDataChunkSetInfo(chunk, time, quality);
    
    LockWrite(ctx->lock);
    err = opcDataRepresentationImportData(rctx, chunk, data, size);
	// ToDo: prelock datachunk for all subscriptions may need it
    LockReturn(ctx->lock);

    if (err) {
	opcChunkPoolReturnChunk(ctx->chunk_pool, chunk);
	return err;
    }
    
    LockWrite(ctx->chunk_lock);
	// ToDo: latest chunk? If we are older than current one? Slow network issues?
    if (ctx->current) {
	ctx->current->next = chunk;
	ctx->current=chunk;
    } else {
	ctx->chunks = chunk;
	ctx->current = chunk;
    }
    LockReturn(ctx->chunk_lock);
    
    return 0;
}

int opcDataPushNative(OPCDataContext ctx, const OPCTime time, const OPCQuality quality, const void *data, unsigned long size) {
    return opcDataPush(ctx, OPC_REPRESENTATION_NATIVE, time, quality, data, size);
}

int opcDataPushOPCXML(OPCDataContext ctx, const OPCTime time, const OPCQuality quality, const void *data, unsigned long size) {
    return opcDataPush(ctx, OPC_REPRESENTATION_OPCXML, time, quality, data, size);
}

int opcDataGet(OPCDataContext ctx, unsigned short int representation, OPCDataInfo *info, OPCData *data) {
    int err;
    OPCDataChunkContext current;

    assert(ctx);
    assert(ctx->current);
    assert(ctx->representation);
    if ((!ctx->current)||(!ctx->representation)) return ERR_OPC_DATA_NOTINITIALISED;
    if (representation>=ctx->supported_representations) return ERR_OPC_DATA_INVALID_REPRESENTATION; 

    LockRead(ctx->chunk_lock);
    current = ctx->current;
    opcDataChunkRef(current);
    LockReturn(ctx->chunk_lock);

    err=opcDataRepresentationGet(ctx->representation[representation], current);

    if (err) {
	if (data) *data=NULL;
    } else {
	if (data) *data=opcDataChunkGet(current, representation);
	if (info) *info=opcDataChunkGetInfo(current);
    }
    
    return err;
}

int opcDataGetNative(OPCDataContext ctx, OPCDataInfo *info, OPCData *data) {
    return opcDataGet(ctx, OPC_REPRESENTATION_NATIVE, info, data);
}

int opcDataGetOPCXML(OPCDataContext ctx, OPCDataInfo *info, OPCData *data) {
    return opcDataGet(ctx, OPC_REPRESENTATION_OPCXML, info, data);
}

void opcDataReturn(OPCDataContext ctx, OPCDataInfo info) {
    OPCDataChunkContext chunk, nchunk;
    unsigned long ref;

    assert(ctx);

	/* ToDo: Minimal Chunks to store > 1 */
	/* After chunk_lock locked number of referers can decrease, but not increase! */
    LockWrite(ctx->chunk_lock);
    ref=opcDataChunkUnrefByInfo(info);
    if (!ref) {
	chunk = opcDataChunkGetChunkByInfo(info);
	nchunk = chunk->next;
	if ((nchunk)&&(ctx->chunks==chunk)) {
	    ctx->chunks=nchunk;
	    LockReturn(ctx->chunk_lock);
	    opcChunkPoolReturnChunk(ctx->chunk_pool, chunk);
	    LockWrite(ctx->chunk_lock);
	    for (chunk=nchunk,nchunk=chunk->next;((nchunk)&&(!opcDataChunkGetRefs(chunk)));chunk=nchunk, nchunk=chunk->next) {
		ctx->chunks=nchunk;
		LockReturn(ctx->chunk_lock);
	    	opcChunkPoolReturnChunk(ctx->chunk_pool, chunk);
		LockWrite(ctx->chunk_lock);
	    }
	    ctx->chunks = chunk;
	}
    }
    LockReturn(ctx->chunk_lock);
}
