
#include <opclib.h>
#include <dsassert.h>
#include <dsdebug.h>

#include <opcformats.h>
#include <opcformatnative.h>
#include <opcformatopcxml.h>

void opcFormatsZeroContext(OPCFormatsContext ctx) {
    ctx->formats = NULL;
    ctx->binary_format = NULL;
    ctx->native_format = NULL;
    ctx->opcxml_format = NULL;
}

OPCFormatsContext opcFormatsCreateContext(DSMallocContext mctx) {
    OPCFormatsContext ctx;

    dsDebugServer(DS_DEBUG_NOTICE, "opcFormatsCreateContext");

    ctx = (OPCFormatsContext)dsMalloc(mctx,sizeof(struct OPCFormatsContextT));
    if (ctx) {
	ctx->malloc=mctx;
	opcFormatsZeroContext(ctx);
    }
    return ctx;
}

int opcFormatsInitContext(OPCFormatsContext ctx, OPCConfigurationContext config) {
    unsigned long formats_hash_size;

    ds_assert_param(ctx);
    ds_assert_param(config);
    
    if ((!ctx)||(!config)) return DS_ERR_INVALID_ARGUMENT;

    formats_hash_size = config->formats_hash_size;
    if (formats_hash_size<4) formats_hash_size=4;
    
    dsDebugServer(DS_DEBUG_NOTICE, "opcFormatsInitContext");
    opcFormatsZeroContext(ctx);

    ctx->formats = dsHashTableInit(ctx->malloc, formats_hash_size, DS_HASH_TYPE_POINTER, NULL);
    if (!ctx->formats) {
	opcFormatsFreeContext(ctx);
	return DS_ERR_MEM;
    }
    dsHashTableSetDestructor(ctx->formats, (HashTableDestructorFunction)opcFormatDestroyContext);

    ctx->native_format = (OPCFormatNativeContext)opcFormatsRegisterFormat(ctx, opcFormatNative(), NULL);
    if (!ctx->native_format) {
	opcFormatsFreeContext(ctx);
	return OPC_ERR_FORMATS_REGISTER_NATIVE;
    }

    ctx->opcxml_format = opcFormatsRegisterFormat(ctx, opcFormatOPCXML(), NULL);
    if (!ctx->opcxml_format) {
	opcFormatsFreeContext(ctx);
	return OPC_ERR_FORMATS_REGISTER_OPCXML;
    }
  
    return 0;
}

void opcFormatsFreeContext(OPCFormatsContext ctx) {
    ds_assert_return(ctx);
    dsDebugServer(DS_DEBUG_NOTICE, "opcFormatsFreeContext");
    
    if (ctx->formats) dsHashTableDestroy(ctx->formats);
    opcFormatsZeroContext(ctx);
}

void opcFormatsDestroyContext(OPCFormatsContext ctx) {
    dsDebugServer(DS_DEBUG_NOTICE, "opcFormatsDestroyContext");
    opcFormatsFreeContext(ctx);
    dsFree(ctx->malloc,ctx);
}


OPCFormatContext opcFormatsFindFormat(OPCFormatsContext ctx, const char *format) {
    ds_assert_null(ctx);
    ds_assert_null(format);
    
    return (OPCFormatContext)dsHashTableFind(ctx->formats, (void*)format);
}


OPCFormatContext opcFormatsRegisterFormat(OPCFormatsContext ctx, OPCFormatContext fctx, const void *options) {
    int err;
    OPCFormatContext format;
    
    ds_assert_null(ctx);
    ds_assert_null(fctx);
    
    format = opcFormatsFindFormat(ctx, fctx->title);
    if (format) {
//	dsDebugServer()
	return format;
    }
    
    format = opcFormatCreateContext(fctx, ctx->malloc);
    if (!format) {
//	dsDebugServer()
	return NULL;
    }
    
    err = opcFormatInitContext(format, options);
    if (err) {
	opcFormatDestroyContext(format);
//	dsDebugServer()
	return NULL;
    }

    if (dsHashTableInsert(ctx->formats, format)) {
	opcFormatDestroyContext(format);
//	dsDebugServer()
	return NULL;
    }
    
    return format;
}
