#define _DS_INCLUDE_XML_HEADERS

#include <stdio.h>
#include <dsassert.h>

#include <libxml/schemasInternals.h>

#include <dslib.h>
#include <dsdebug.h>

#include <opcfdopcbinarydictionary.h>
#include <opcfdopcbinary.h>

void opcFDOPCBinaryZeroContext(OPCFDOPCBinaryContext ctx) {
    ds_assert_return(ctx);
    
    ctx->lock = NULL;
    ctx->dictdir = NULL;
    ctx->schema_rep = NULL;
}

OPCFDOPCBinaryContext opcFDOPCBinaryCreateContext(DSMallocContext mctx) {
    OPCFDOPCBinaryContext ctx;

    dsDebugServer(DS_DEBUG_NOTICE, "opcFormatDescriptionOPCBinaryCreateContext");

    ctx = (OPCFDOPCBinaryContext)dsMalloc(mctx,sizeof(struct OPCFDOPCBinaryContextT));
    if (ctx) {
	ctx->malloc=mctx;
	opcFDOPCBinaryZeroContext(ctx);
    }
    return ctx;
}

int opcFDOPCBinaryInitContext(OPCFDOPCBinaryContext ctx, OPCNamespaces namespaces, const char *schema, const char *dictdir) {
    int err;
    xmlSchemaParserCtxtPtr schema_parser;

    ds_assert_param(ctx);
    ds_assert_param(namespaces);
    
    dsDebugServer(DS_DEBUG_NOTICE, "opcFormatDescriptionOPCBinaryInitContext: %s", schema);

    opcFDOPCBinaryZeroContext(ctx);

    if (dsExchangeString(ctx->malloc, &ctx->dictdir, dictdir)) return DS_ERR_MEM;
    ctx->dictdir_chars = strlen(ctx->dictdir);

    schema_parser=xmlSchemaNewParserCtxt(schema);
    if (!schema_parser) {
	opcFDOPCBinaryFreeContext(ctx);
	return OPC_ERR_FD_OPCBINARY_PARSE;
    }
    ctx->schema_rep=xmlSchemaParse(schema_parser);
    xmlSchemaFreeParserCtxt(schema_parser);
    if (!ctx->schema_rep) {
	opcFDOPCBinaryFreeContext(ctx);
	return OPC_ERR_FD_OPCBINARY_PARSE;
    }
    
    ctx->lock = dsLockCreateContext(ctx->malloc);
    if (!ctx->lock) {
	opcFDOPCBinaryFreeContext(ctx);
	return DS_ERR_MEM;
    }
    err = dsLockInitContext(ctx->lock);
    if (err) {
	opcFDOPCBinaryFreeContext(ctx);
	return err;
    }

    err=opcNamespacesSetOPCBinaryNamespace(namespaces, ctx->schema_rep->targetNamespace);
    if (err) {
	opcFDOPCBinaryFreeContext(ctx);
	return err;
    }
    ctx->namespaces = namespaces;

//    xmlSchemaSetValidErrors(ctx->schema_ctx,NULL,NULL,NULL);

    return 0;        
}

void opcFDOPCBinaryFreeContext(OPCFDOPCBinaryContext ctx) {
    ds_assert_return(ctx);

    dsDebugServer(DS_DEBUG_NOTICE, "opcFormatDescriptionOPCBinaryFreeContext");

    if (ctx->lock) dsLockDestroyContext(ctx->lock);
    if (ctx->dictdir) dsExchangeString(ctx->malloc, &ctx->dictdir, NULL);
    if (ctx->schema_rep) xmlSchemaFree(ctx->schema_rep);
    
    opcFDOPCBinaryZeroContext(ctx);
}

void opcFDOPCBinaryDestroyContext(OPCFDOPCBinaryContext ctx) {
    dsDebugServer(DS_DEBUG_NOTICE, "opcFormatDescriptionOPCBinaryDestroyContext");
    opcFDOPCBinaryFreeContext(ctx);
    dsFree(ctx->malloc,ctx);
}

int opcFDOPCBinaryValidateXML(OPCFDOPCBinaryContext ctx, xmlDocPtr doc) {
    return 0;
}

int opcFDOPCBinaryValidateDictionary(OPCFDOPCBinaryContext ctx, OPCFDOPCBinaryDictionaryContext dctx) {
    dsLockWrite(ctx->lock);
    dsLockReturn(ctx->lock);
    return 0;
}

OPCFDOPCBinaryDictionaryContext opcFDOPCBinaryCreateDictionaryFromMemory(OPCFDOPCBinaryContext ctx, const char *xml, unsigned long length) {
    int err;
    xmlDocPtr doc;
    xmlSchemaValidCtxtPtr schema_ctx;
    OPCFDOPCBinaryDictionaryContext dctx;
    
    ds_assert_null(ctx);
    ds_assert_null(xml);
    ds_assert_null(length>0);
    ds_assert_null(ctx->schema_rep);

    dsDebugServer(DS_DEBUG_NOTICE, "opcFormatDescriptionOPCBinaryCreateDictionaryFromMemory");

    if (!length) length=xmlStrlen(xml);
    doc = xmlReadMemory(xml,length,"xml",NULL,0);
    if (!doc) return NULL;
    
    dsLockWrite(ctx->lock);
    schema_ctx = xmlSchemaNewValidCtxt(ctx->schema_rep);
    dsLockReturn(ctx->lock);

    if (!schema_ctx) {
	xmlFreeDoc(doc);
	return NULL;
    }

    xmlSchemaSetValidErrors(schema_ctx,NULL,NULL,NULL);
    err = xmlSchemaValidateDoc(schema_ctx,doc);
    xmlSchemaFreeValidCtxt(schema_ctx);
    
    if (err) {
	xmlFreeDoc(doc);
	return NULL;
    }

    dctx=opcFDOPCBinaryDictionaryCreateContext(ctx->malloc);
    if (!dctx) return dctx;
    
    err = opcFDOPCBinaryDictionaryInitContext(dctx,ctx->namespaces,doc);
    if (err) {
	xmlFreeDoc(doc);
	opcFDOPCBinaryDictionaryDestroyContext(dctx);
	return NULL;
    }

    dsDebugServer(DS_DEBUG_NOTICE, "opcFormatDescriptionOPCBinaryCreateDictionary: Ok. Created %p", dctx);

    return dctx;    
}

OPCFDOPCBinaryDictionaryContext opcFDOPCBinaryCreateDictionary(OPCFDOPCBinaryContext ctx, const char *dictionary) {
    int err;
    char *stmp;
    xmlDocPtr doc;
    xmlSchemaValidCtxtPtr schema_ctx;
    OPCFDOPCBinaryDictionaryContext dctx;
    
    ds_assert_null(ctx);
    ds_assert_null(dictionary);
    ds_assert_null(ctx->schema_rep);

    dsDebugServer(DS_DEBUG_NOTICE, "opcFormatDescriptionOPCBinaryCreateDictionary: %s", dictionary);

    if (dictionary[0]=='/') doc = xmlReadFile(dictionary,NULL,0);
    else {
	stmp = (char*)dsMalloca(ctx->dictdir_chars+strlen(dictionary)+2);
	if (!stmp) return NULL;
	sprintf(stmp,"%s/%s",ctx->dictdir,dictionary);
	doc = xmlReadFile(stmp,NULL,0);
	dsFreea(stmp);
    }
    if (!doc) return NULL;

    
    dsLockWrite(ctx->lock);
    schema_ctx = xmlSchemaNewValidCtxt(ctx->schema_rep);
    dsLockReturn(ctx->lock);

    if (!schema_ctx) {
	xmlFreeDoc(doc);
	return NULL;
    }

    xmlSchemaSetValidErrors(schema_ctx,NULL,NULL,NULL);
    err = xmlSchemaValidateDoc(schema_ctx,doc);
    xmlSchemaFreeValidCtxt(schema_ctx);
    
    if (err) {
	xmlFreeDoc(doc);
	return NULL;
    }

    dctx=opcFDOPCBinaryDictionaryCreateContext(ctx->malloc);
    if (!dctx) return dctx;
    
    err = opcFDOPCBinaryDictionaryInitContext(dctx,ctx->namespaces,doc);
    if (err) {
	xmlFreeDoc(doc);
	opcFDOPCBinaryDictionaryDestroyContext(dctx);
	return NULL;
    }

    dsDebugServer(DS_DEBUG_NOTICE, "opcFormatDescriptionOPCBinaryCreateDictionary: Ok. Created %p", dctx);

    return dctx;    
}
